@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import base64
@@ -6,20 +6,15 @@ import json
 import os
 import re
 import sys
+from collections.abc import Iterable, Iterator, Sequence
 from contextlib import suppress
 from functools import lru_cache, partial
-from time import monotonic, time_ns
+from time import time_ns
 from types import GeneratorType
 from typing import (
     TYPE_CHECKING,
     Any,
-    Dict,
-    FrozenSet,
-    Iterable,
-    Iterator,
-    List,
     Optional,
-    Tuple,
     Union,
     cast,
 )
@@ -33,6 +28,7 @@ from .fast_data_types import (
     EllipticCurveKey,
     get_boss,
     get_options,
+    monotonic,
     read_command_response,
     send_data_to_peer,
 )
@@ -41,8 +37,8 @@ from .types import AsyncResponse
 from .typing import BossType, WindowType
 from .utils import TTYIO, log_error, parse_address_spec, resolve_custom_file
 
-active_async_requests: Dict[str, float] = {}
-active_streams: Dict[str, str] = {}
+active_async_requests: dict[str, float] = {}
+active_streams: dict[str, str] = {}
 if TYPE_CHECKING:
     from .window import Window
 
@@ -51,12 +47,16 @@ def encode_response_for_peer(response: Any) -> bytes:
     return b'\x1bP@kitty-cmd' + json.dumps(response).encode('utf-8') + b'\x1b\\'
 
 
-def parse_cmd(serialized_cmd: str, encryption_key: EllipticCurveKey) -> Dict[str, Any]:
+def parse_cmd(serialized_cmd: memoryview, encryption_key: EllipticCurveKey) -> dict[str, Any]:
+    # See https://github.com/python/cpython/issues/74379 for why we cant use
+    # memoryview directly :((
     try:
-        pcmd = json.loads(serialized_cmd)
+        pcmd = json.loads(bytes(serialized_cmd))
     except Exception:
+        log_error('Failed to parse JSON payload of remote command, ignoring it')
         return {}
     if not isinstance(pcmd, dict) or 'version' not in pcmd:
+        log_error('JSON payload of remote command is invalid, must be an object with a version field')
         return {}
     pcmd.pop('password', None)
     if 'encrypted' in pcmd:
@@ -82,7 +82,7 @@ def parse_cmd(serialized_cmd: str, encryption_key: EllipticCurveKey) -> Dict[str
 
 class CMDChecker:
 
-    def __call__(self, pcmd: Dict[str, Any], window: Optional['Window'], from_socket: bool, extra_data: Dict[str, Any]) -> Optional[bool]:
+    def __call__(self, pcmd: dict[str, Any], window: Optional['Window'], from_socket: bool, extra_data: dict[str, Any]) -> Optional[bool]:
         return False
 
 
@@ -104,9 +104,30 @@ def fnmatch_pattern(pat: str) -> 're.Pattern[str]':
     return re.compile(translate(pat))
 
 
+def remote_control_allowed(
+    pcmd: dict[str, Any], remote_control_passwords: Optional[dict[str, Sequence[str]]],
+    window: Optional['Window'], extra_data: dict[str, Any]
+) -> bool:
+    if not remote_control_passwords:
+        return True
+    pw = pcmd.get('password', '')
+    auth_items = remote_control_passwords.get(pw)
+    if pw == '!':
+        auth_items = None
+    if auth_items is None:
+        if '!' in remote_control_passwords:
+            raise PermissionError()
+        return False
+    from .remote_control import password_authorizer
+    pa = password_authorizer(auth_items)
+    if not pa.is_cmd_allowed(pcmd, window, False, extra_data):
+        raise PermissionError()
+    return True
+
+
 class PasswordAuthorizer:
 
-    def __init__(self, auth_items: FrozenSet[str]) -> None:
+    def __init__(self, auth_items: frozenset[str]) -> None:
         self.command_patterns = []
         self.function_checkers = []
         self.name = ''
@@ -117,7 +138,7 @@ class PasswordAuthorizer:
             else:
                 self.command_patterns.append(fnmatch_pattern(item))
 
-    def is_cmd_allowed(self, pcmd: Dict[str, Any], window: Optional['Window'], from_socket: bool, extra_data: Dict[str, Any]) -> bool:
+    def is_cmd_allowed(self, pcmd: dict[str, Any], window: Optional['Window'], from_socket: bool, extra_data: dict[str, Any]) -> bool:
         cmd_name = pcmd.get('cmd')
         if not cmd_name:
             return False
@@ -140,14 +161,14 @@ class PasswordAuthorizer:
 
 
 @lru_cache(maxsize=256)
-def password_authorizer(auth_items: FrozenSet[str]) -> PasswordAuthorizer:
+def password_authorizer(auth_items: frozenset[str]) -> PasswordAuthorizer:
     return PasswordAuthorizer(auth_items)
 
 
-user_password_allowed: Dict[str, bool] = {}
+user_password_allowed: dict[str, bool] = {}
 
 
-def is_cmd_allowed(pcmd: Dict[str, Any], window: Optional['Window'], from_socket: bool, extra_data: Dict[str, Any]) -> Optional[bool]:
+def is_cmd_allowed(pcmd: dict[str, Any], window: Optional['Window'], from_socket: bool, extra_data: dict[str, Any]) -> Optional[bool]:
     sid = pcmd.get('stream_id', '')
     if sid and active_streams.get(sid, '') == pcmd['cmd']:
         return True
@@ -184,8 +205,8 @@ def close_active_stream(stream_id: str) -> None:
 
 
 def handle_cmd(
-    boss: BossType, window: Optional[WindowType], cmd: Dict[str, Any], peer_id: int, self_window: Optional[WindowType]
-) -> Union[Dict[str, Any], None, AsyncResponse]:
+    boss: BossType, window: Optional[WindowType], cmd: dict[str, Any], peer_id: int, self_window: Optional[WindowType]
+) -> Union[dict[str, Any], None, AsyncResponse]:
     v = cmd['version']
     no_response = cmd.get('no_response', False)
     if tuple(v)[:2] > version[:2]:
@@ -228,7 +249,7 @@ def handle_cmd(
         if stream:
             return {'ok': True, 'stream': True}
         return ans
-    response: Dict[str, Any] = {'ok': True}
+    response: dict[str, Any] = {'ok': True}
     if ans is not None:
         response['data'] = ans
     if not no_response:
@@ -249,7 +270,8 @@ they will only work if this process is run within a kitty window.
 --password
 A password to use when contacting kitty. This will cause kitty to ask the user
 for permission to perform the specified action, unless the password has been
-accepted before or is pre-configured in :file:`kitty.conf`.
+accepted before or is pre-configured in :file:`kitty.conf`. To use a blank password
+specify :option:`kitten @ --use-password` as :code:`always`.
 
 
 --password-file
@@ -257,14 +279,14 @@ completion=type:file relative:conf kwds:-
 default=rc-pass
 A file from which to read the password. Trailing whitespace is ignored. Relative
 paths are resolved from the kitty configuration directory. Use - to read from STDIN.
-Used if no :option:`kitty @ --password` is supplied. Defaults to checking for the
+Used if no :option:`kitten @ --password` is supplied. Defaults to checking for the
 :file:`rc-pass` file in the kitty configuration directory.
 
 
 --password-env
 default=KITTY_RC_PASSWORD
 The name of an environment variable to read the password from.
-Used if no :option:`kitty @ --password-file` is supplied. Defaults
+Used if no :option:`kitten @ --password-file` is supplied. Defaults
 to checking the environment variable :envvar:`KITTY_RC_PASSWORD`.
 
 
@@ -273,7 +295,7 @@ default=if-available
 choices=if-available,never,always
 If no password is available, kitty will usually just send the remote control command
 without a password. This option can be used to force it to :code:`always` or :code:`never` use
-the supplied password.
+the supplied password. If set to always and no password is provided, the blank password is used.
 '''.format, appname=appname)
 
 
@@ -333,14 +355,14 @@ class SocketIO:
 class RCIO(TTYIO):
 
     def simple_recv(self, timeout: float) -> bytes:
-        ans: List[bytes] = []
+        ans: list[bytes] = []
         read_command_response(self.tty_fd, timeout, ans)
         return b''.join(ans)
 
 
 def do_io(
-    to: Optional[str], original_cmd: Dict[str, Any], no_response: bool, response_timeout: float, encrypter: 'CommandEncrypter'
-) -> Dict[str, Any]:
+    to: Optional[str], original_cmd: dict[str, Any], no_response: bool, response_timeout: float, encrypter: 'CommandEncrypter'
+) -> dict[str, Any]:
     payload = original_cmd.get('payload')
     if not isinstance(payload, GeneratorType):
         send_data: Union[bytes, Iterator[bytes]] = encode_send(encrypter(original_cmd))
@@ -359,7 +381,7 @@ def do_io(
             return {'ok': True}
         received = io.simple_recv(timeout=response_timeout)
 
-    return cast(Dict[str, Any], json.loads(received.decode('ascii')))
+    return cast(dict[str, Any], json.loads(received.decode('ascii')))
 
 
 cli_msg = (
@@ -369,7 +391,7 @@ cli_msg = (
 ).format(appname=appname)
 
 
-def parse_rc_args(args: List[str]) -> Tuple[RCOptions, List[str]]:
+def parse_rc_args(args: list[str]) -> tuple[RCOptions, list[str]]:
     cmap = {name: command_for_name(name) for name in sorted(all_command_names())}
     cmds = (f'  :green:`{cmd.name}`\n    {cmd.short_desc}' for c, cmd in cmap.items())
     msg = cli_msg + (
@@ -394,7 +416,7 @@ class CommandEncrypter:
         self.encryption_version = encryption_version
         self.password = password
 
-    def __call__(self, cmd: Dict[str, Any]) -> Dict[str, Any]:
+    def __call__(self, cmd: dict[str, Any]) -> dict[str, Any]:
         encrypter = AES256GCMEncrypt(self.secret)
         cmd['timestamp'] = time_ns()
         cmd['password'] = self.password
@@ -418,14 +440,14 @@ class NoEncryption(CommandEncrypter):
 
     def __init__(self) -> None: ...
 
-    def __call__(self, cmd: Dict[str, Any]) -> Dict[str, Any]:
+    def __call__(self, cmd: dict[str, Any]) -> dict[str, Any]:
         return cmd
 
     def adjust_response_timeout_for_password(self, response_timeout: float) -> float:
         return response_timeout
 
 
-def create_basic_command(name: str, payload: Any = None, no_response: bool = False, is_asynchronous: bool = False) -> Dict[str, Any]:
+def create_basic_command(name: str, payload: Any = None, no_response: bool = False, is_asynchronous: bool = False) -> dict[str, Any]:
     ans = {'cmd': name, 'version': version, 'no_response': no_response}
     if payload is not None:
         ans['payload'] = payload
@@ -439,7 +461,7 @@ def send_response_to_client(data: Any = None, error: str = '', peer_id: int = 0,
     if active_async_requests.pop(async_id, None) is None:
         return
     if error:
-        response: Dict[str, Union[bool, int, str]] = {'ok': False, 'error': error}
+        response: dict[str, Union[bool, int, str]] = {'ok': False, 'error': error}
     else:
         response = {'ok': True, 'data': data}
     if peer_id > 0:
@@ -485,7 +507,7 @@ def get_password(opts: RCOptions) -> str:
     return ans
 
 
-def get_pubkey() -> Tuple[str, bytes]:
+def get_pubkey() -> tuple[str, bytes]:
     raw = os.environ.get('KITTY_PUBLIC_KEY', '')
     if not raw:
         raise SystemExit('Password usage requested but KITTY_PUBLIC_KEY environment variable is not available')
